#include <iostream>
#include <array>
#include <cmath>
using namespace std; 

enum Coord {X, Y};
constexpr double max_delta_tr(5.);  // déplacement max en translation par mise à jour
constexpr double max_delta_rt(0.8); // déplacement max en rotation (rd) par mise à jour

typedef double Orient;       // en rd
typedef array<double,2> S2d; // utilisable pour point et vecteur

double s2d_norm(S2d tab);
double s2d_prod_scal(S2d v1, S2d v2);
void   s2d_add_scaled_vector(S2d& pos, const S2d& pos_to_goal, double scaling);

class Rob 
{ 
public:
	Rob(S2d pos={0.,0.}):pos(pos){}
	void move_to(S2d goal) {pos=goal;} // téléportation
	void affiche() {cout << pos[X] << "\t" << pos[Y] << endl;}
protected: 
	S2d pos; 
}; 

class Omni: public Rob // robot Omnidirectionnel, pas besoin de tourner
{ 
public: 
	Omni ():Rob(){}
	void move_to(S2d goal); // avec limitation en translation
}; 

// se déplace au plus de max_delta_tr vers le point goal
void Omni::move_to(S2d goal)
{
	S2d pos_to_goal = {goal[X] - pos[X], goal[Y] - pos[Y]} ;
	double norm(s2d_norm(pos_to_goal));
	
	if(norm <= max_delta_tr) pos = goal;
	else s2d_add_scaled_vector(pos, pos_to_goal, max_delta_tr/norm);
}

class Dir: public Rob // robot directionnel, doit tourner
{ 
public: 
	Dir():Rob(),a(0){}
	void move_to(S2d goal); // avance puis tourne (avec limitations)
private:
	Orient a; // angle représentation l'orientation en rd
}; 

// se déplace selon son orientation courante avec limitation supplémentaire
// puis s'oriente au plus de max_delta_rt en direction de updated_pos_to_goal
void Dir::move_to(S2d goal)
{
	// mise à jour de la position avec un déplacement selon l'orientation courante
	S2d    init_pos_to_goal = {goal[X] - pos[X], goal[Y] - pos[Y]};
	S2d    travel_dir    = {cos(a), sin(a)}; //vecteur unitaire selon Xrobot
	double proj_goal= s2d_prod_scal(init_pos_to_goal, travel_dir);

    if(abs(proj_goal) > max_delta_tr) 
		proj_goal = ((proj_goal > 0) ? 1 : -1)*max_delta_tr;
    
	s2d_add_scaled_vector(pos, travel_dir, proj_goal);
	
	// mise à jour de l'orientation
	S2d    updated_pos_to_goal = {goal[X] - pos[X], goal[Y] - pos[Y]};
	Orient goal_a(atan2(updated_pos_to_goal[Y],updated_pos_to_goal[X]));
	Orient delta_a(goal_a - a);
	
	if(abs(delta_a) <= max_delta_rt) a = goal_a ; 
	else  a += ((delta_a > 0) ?  1. : -1.)*max_delta_rt ;
}

int main()
{   // les 3 robots sont initialisés à l'origine (0,0)
	Rob a;
	Omni b;
	Dir c;
	
	// téléportation
	a.affiche();
	a.move_to({10., 10.});
	a.affiche();
	cout << endl;
	
	// omnidirectionnel
	b.affiche();
	for(int i(0) ; i<3 ; ++i)
	{
		b.move_to({10., 10.});
		b.affiche();
	}
	cout << endl;
	
	// non-holonome
	c.affiche();
	for(int i(0) ; i<4 ; ++i)
	{
		c.move_to({10., 10.});
		c.affiche();
	}
	cout << endl;
	return 0;
}

double s2d_norm(S2d tab)
{
	return sqrt(tab[X]*tab[X] + tab[Y]*tab[Y]);
}

double s2d_prod_scal(S2d v1, S2d v2)
{
	return v1[X]*v2[X] + v1[Y]*v2[Y];
}

void s2d_add_scaled_vector(S2d& pos, const S2d& pos_to_goal, double scaling)
{
	pos[X] += scaling*pos_to_goal[X];
	pos[Y] += scaling*pos_to_goal[Y];
}
